// SPDX-License-Identifier: LGPL-2.1
// Copyright The Music Player Daemon Project
// Based on AudioCompress (c)2007 busybee (http://beesbuzz.biz/

#include "Normalizer.hxx"

/*** Default configuration stuff ***/
#define TARGET 16384		/*!< Target level (on a scale of 0-32767) */

#define GAINMAX 32		/*!< The maximum amount to amplify by */
#define GAINSMOOTH 8		/*!< How much inertia ramping has*/
#define BUCKETS 400		/*!< How long of a history to use by default */

struct Compressor {
	int target;
	int maxgain;
	int smooth;

        //! History of the peak values
        int *peaks;

        //! History of the gain values
        int *gain;

        //! History of clip amounts
        int *clipped;

        unsigned int pos;
        unsigned int bufsz;
};

struct Compressor *
Compressor_new(unsigned int history) noexcept
{
	auto obj = new Compressor();

	obj->target = TARGET;
	obj->maxgain = GAINMAX;
	obj->smooth = GAINSMOOTH;

        obj->peaks = obj->gain = obj->clipped = nullptr;
	obj->bufsz = 0;
        obj->pos = 0;

	if (!history)
                history = BUCKETS;

        obj->bufsz = history;
        obj->peaks = new int[history]{};
        obj->gain = new int[history]{};
        obj->clipped = new int[history]{};

        return obj;
}

void
Compressor_delete(struct Compressor *obj) noexcept
{
	delete[] obj->peaks;
	delete[] obj->gain;
	delete[] obj->clipped;
	delete obj;
}

void
Compressor_Process_int16(struct Compressor *obj, int16_t *audio,
			 unsigned int count) noexcept
{
	int16_t *ap;
	unsigned int i;
        int *peaks = obj->peaks;
        int curGain = obj->gain[obj->pos];
        int newGain;
        int peakVal = 1;
        int peakPos = 0;
        int slot = (obj->pos + 1) % obj->bufsz;
        int *clipped = obj->clipped + slot;
        unsigned int ramp = count;
        int delta;

	ap = audio;
	for (i = 0; i < count; i++)
	{
		int val = *ap++;
                if (val < 0)
                        val = -val;
		if (val > peakVal)
                {
			peakVal = val;
                        peakPos = i;
                }
	}
	peaks[slot] = peakVal;


	for (i = 0; i < obj->bufsz; i++)
	{
		if (peaks[i] > peakVal)
		{
			peakVal = peaks[i];
			peakPos = 0;
		}
	}

	//! Determine target gain
	newGain = (1 << 10)*obj->target/peakVal;

        //! Adjust the gain with inertia from the previous gain value
        newGain = (curGain*((1 << obj->smooth) - 1) + newGain)
                >> obj->smooth;

        //! Make sure it's no more than the maximum gain value
        if (newGain > (obj->maxgain << 10))
                newGain = obj->maxgain << 10;

        //! Make sure it's no less than 1:1
	if (newGain < (1 << 10))
		newGain = 1 << 10;

        //! Make sure the adjusted gain won't cause clipping
        if ((peakVal*newGain >> 10) > 32767)
        {
                newGain = (32767 << 10)/peakVal;
                //! Truncate the ramp time
                ramp = peakPos;
        }

        //! Record the new gain
        obj->gain[slot] = newGain;

        if (!ramp)
                ramp = 1;
        if (!curGain)
                curGain = 1 << 10;
	delta = (newGain - curGain) / (int)ramp;

	ap = audio;
        *clipped = 0;
	for (i = 0; i < count; i++)
	{
		int sample;

		//! Amplify the sample
		sample = *ap*curGain >> 10;
		if (sample < -32768)
		{
			*clipped += -32768 - sample;
			sample = -32768;
		} else if (sample > 32767)
		{
			*clipped += sample - 32767;
			sample = 32767;
		}
		*ap++ = sample;

                //! Adjust the gain
                if (i < ramp)
                        curGain += delta;
                else
                        curGain = newGain;
	}

        obj->pos = slot;
}

